
{** @abstract(Arbitre du jeu d'échecs.)
  Arbitre du jeu d'échecs. }
unit chessgame;
{$IFDEF FPC}
{$MODE DELPHI}
{$ENDIF}

interface

uses
  classes;

const
  cNil = 0;
  cPawn = 1;
  cKnight = 2;
  cBishop = 3;
  cRook = 4;
  cQueen = 5;
  cKing = 6;
  cBorder = 7;
  cWhite = 1;
  cBlack = -1;
  cNone = 0;
  
  cDefaultPosition = 'rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1';

type
  TBoard = array[-1..10, -1..10] of integer;

  TCastling = (e1g1, e1c1, e8g8, e8c8);

  TPosition = class
    private
      board: TBoard;
      activeColor: integer;
      castling: string[4];
      enPassant: string[2];
      halfMoves: integer;
      fullMove: integer;
      list: TStringList;
      check1: boolean;
      castlingcheck1: array[e1g1..e8c8] of boolean;
    public
      constructor Create;
      destructor Destroy; override;
      procedure Initialize(const s: string = '');
      function BoardAsText: string;
      procedure GenerateMoves(const aColor: integer; const aAllMoves: boolean);
      procedure GenerateCastling(const c: char);
      function Check: boolean;
      function CastlingCheck(c: char): boolean;
      procedure CastlingRemove(const c: char);
      procedure PlayMove(const aMove: string; const aPromotion: integer);
      function ForsythEdwards: string;
      procedure UpdateVars;
  end;

  TState = (progress, checkmate, stalemate, draw);

  TGame = class
    private
      current, next: TPosition;
      list: TStringList;
      state: TState;
    public
      constructor Create; overload;
      constructor Create(const s: string); overload;
      destructor Destroy; override;
      procedure GenerateLegalMoves;
      function IsLegal(const s: string): boolean;
      function ActiveColor: integer;
      function Check: boolean;
      function ForsythEdwards: string;
      procedure PlayMove(const aMove: string; const aPromotion: integer);
      function CurrentBoardAsText: string;
      
      function IsPromotion(const aMove: string): boolean;
      function IsCastling(const aMove: string): string;
      function IsEnPassant(const aMove: string): string;
      
      property currentstate: TState read state;
  end;

const
  cStateStr: array[TState] of string = (
    'progress',
    'checkmate',
    'stalemate',
    'draw'
    );

implementation

uses
  sysutils,

  chessgameutils;

var
  loglist: TStringList;
  logfile: string;

procedure ToLog(const aText: string);
begin
  loglist.Add(Format('%s %s', [TimeToStr(Now), aText]));
end;

constructor TPosition.Create;
begin
  inherited Create;
  list := TStringList.Create;
end;

destructor TPosition.Destroy;
begin
  list.Free;
  inherited Destroy;
end;

procedure TPosition.Initialize(const s: string = '');
const
  cBoard: TBoard = (
    (7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7),
    (7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7),
    (7, 7, 4, 1, 0, 0, 0, 0, -1, -4, 7, 7),
    (7, 7, 2, 1, 0, 0, 0, 0, -1, -2, 7, 7),
    (7, 7, 3, 1, 0, 0, 0, 0, -1, -3, 7, 7),
    (7, 7, 5, 1, 0, 0, 0, 0, -1, -5, 7, 7),
    (7, 7, 6, 1, 0, 0, 0, 0, -1, -6, 7, 7),
    (7, 7, 3, 1, 0, 0, 0, 0, -1, -3, 7, 7),
    (7, 7, 2, 1, 0, 0, 0, 0, -1, -2, 7, 7),
    (7, 7, 4, 1, 0, 0, 0, 0, -1, -4, 7, 7),
    (7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7),
    (7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7)
    );
  cSymbols: string = 'kqrbnp PNBRQK';
var
  l: TStringList;
  x, y, i: integer;
  c: char;
begin
  if s = '' then
  begin
    board := cBoard;
    activeColor := cWhite;
    castling := 'KQkq';
    enPassant := '-';
    halfMoves := 0;
    fullMove := 1;
  end else
  begin
    l := TStringList.Create;
    with l do
    begin
      DelimitedText := s;
      Assert(Count = 6);
    end;

    for x := -1 to 10 do
      for y := -1 to 10 do
        board[x, y] := cBorder;

    x := 1;
    y := 8;
    i := 1;

    while i <= Length(l[0]) do
    begin
      c := l[0][i];

      case c of
        '/':
          begin
            Dec(y);
            x := 1;
          end;
        '1'..'8':
          while c > '0' do
          begin
            board[x, y] := cNil;
            Inc(x);
            Dec(c);
          end;
      else
        begin
          board[x, y] := Pos(c, cSymbols) - 7;
          Inc(x);
        end;
      end;

      Inc(i);
    end;

    if l[1] = 'w' then activeColor := cWhite else activeColor := cBlack;

    castling := l[2];
    enPassant := l[3];

    Val(l[4], halfMoves, i);
    Val(l[5], fullMove, i);

    l.Free;
  end;
end;

function TPosition.BoardAsText: string;
const
  cSymbols: string = 'kqrbnp-PNBRQK+';
var
  x, y: integer;
begin
  result := '';
  for y := 10 downto -1 do
  begin
    for x := -1 to 10 do
      result := Concat(result, cSymbols[board[x, y] + 7]);
    if y > -1 then
      result := Concat(result, #13#10);
  end;
end;

procedure TPosition.GenerateMoves(const aColor: integer; const aAllMoves: boolean);
var
  x1, y1, x2, y2: integer;
  i, j, k: integer;
begin
  list.Clear;
  for x1 := 1 to 8 do
    for y1 := 1 to 8 do
      if Sign(board[x1, y1]) = aColor then
        case Abs(board[x1, y1]) of
          cPawn:
            begin
              if aColor = cWhite then
              begin
                j := 1;
                k := 2;
              end else
              begin
                j := 3;
                k := 4;
              end;
              for i := j to k do
              begin
                Compute(x2, y2, x1, y1, i);
                if (-1 * aColor * board[x2, y2] in [cPawn..cKing])
                  or (SquareToStr(x2, y2) = string(enPassant)) then
                  list.Add(MoveToStr(x1, y1, x2, y2));
              end;
              if aAllMoves then
              begin
                if aColor = cWhite then
                  i := 7
                else
                  i := 8;
                Compute(x2, y2, x1, y1, i);
                if board[x2, y2] = cNil then
                  list.Add(MoveToStr(x1, y1, x2, y2));
                if (y2 * aColor = 3 * cWhite) or (y2 * aColor = 6 * cBlack) then
                begin
                  Compute(x2, y2, x2, y2, i);
                  if board[x2, y2] = cNil then
                    list.Add(MoveToStr(x1, y1, x2, y2));
                end;
              end;
            end;
          cKnight:
            for i := 9 to 16 do
            begin
              Compute(x2, y2, x1, y1, i);
              if - 1 * aColor * board[x2, y2] in [cNil..cKing] then
                list.Add(MoveToStr(x1, y1, x2, y2));
            end;
          cBishop, cRook, cQueen:
            begin
              case Abs(board[x1, y1]) of
                cBishop:
                  begin
                    j := 1;
                    k := 4;
                  end;
                cRook:
                  begin
                    j := 5;
                    k := 8;
                  end;
                cQueen:
                  begin
                    j := 1;
                    k := 8;
                  end;
              end;
              for i := j to k do
              begin
                x2 := x1;
                y2 := y1;
                repeat
                  Compute(x2, y2, x2, y2, i);
                  if - 1 * aColor * board[x2, y2] in [cNil..cKing] then
                    list.Add(MoveToStr(x1, y1, x2, y2));
                until board[x2, y2] <> cNil;
              end;
            end;
          cKing:
            begin
              for i := 1 to 8 do
              begin
                Compute(x2, y2, x1, y1, i);
                if - 1 * aColor * board[x2, y2] in [cNil..cKing] then
                  list.Add(MoveToStr(x1, y1, x2, y2));
              end;
              if aAllMoves then
              begin
                if activeColor = cWhite then
                begin
                  if Pos('K', castling) > 0 then GenerateCastling('K');
                  if Pos('Q', castling) > 0 then GenerateCastling('Q');
                end else
                begin
                  if Pos('k', castling) > 0 then GenerateCastling('k');
                  if Pos('q', castling) > 0 then GenerateCastling('q');
                end;
              end;
            end;
        end;
end;

procedure TPosition.GenerateCastling(const c: char);
var
  i, j, k, l: integer;
  b: boolean;
  r: TCastling;
begin
  b := true;
  case c of
    'K': begin j := 1; k := 5; l := 8; r := e1g1; end;
    'Q': begin j := 1; k := 1; l := 5; r := e1c1; end;
    'k': begin j := 8; k := 5; l := 8; r := e8g8; end;
    'q': begin j := 8; k := 1; l := 5; r := e8c8; end;
  end;
  for i := k + 1 to l - 1 do
    if board[i, j] <> cNil then
      b := false;
  if b and not castlingcheck1[r] then
  begin
    if k = 5 then
      k := 7
    else
      k := 3;
    list.Add(MoveToStr(5, j, k, j));
  end;
end;

function TPosition.Check: boolean;
var
  i: integer;
  x2, y2: integer;
begin
  result := false;
  for i := 0 to list.count - 1 do
  begin
    x2 := Ord(list[i][3]) - Ord('a') + 1;
    y2 := Ord(list[i][4]) - Ord('1') + 1;
    if board[x2, y2] = activeColor * cKing then
      result := true;
  end;
end;

function TPosition.CastlingCheck(c: char): boolean;
var
  i: integer;
  x2, y2: integer;
  j, k, l: integer;
begin
  case c of
    'K': begin j := 1; k := 5; l := 8; end;
    'Q': begin j := 1; k := 1; l := 5; end;
    'k': begin j := 8; k := 5; l := 8; end;
    'q': begin j := 8; k := 1; l := 5; end;
  end;
  result := false;
  for i := 0 to list.count - 1 do
  begin
    x2 := Ord(list[i][3]) - Ord('a') + 1;
    y2 := Ord(list[i][4]) - Ord('1') + 1;
    if (y2 = j) and (x2 in [k..l]) then
      result := true;
  end;
end;

procedure TPosition.CastlingRemove(const c: char);
var
  i: integer;
begin
  i := Pos(c, castling);
  if i > 0 then
    if Length(castling) = 1 then
      castling := '-'
    else
      Delete(castling, i, 1);
end;

procedure TPosition.PlayMove(const aMove: string; const aPromotion: integer);
var
  x1, y1, x2, y2: integer;
begin
  x1 := Ord(aMove[1]) - Ord('a') + 1;
  y1 := Ord(aMove[2]) - Ord('1') + 1;
  x2 := Ord(aMove[3]) - Ord('a') + 1;
  y2 := Ord(aMove[4]) - Ord('1') + 1;
  if activeColor * board[x1, y1] = cKing then
    case 10 * x1 + y1 of
      51:
        begin
          CastlingRemove('K');
          CastlingRemove('Q');
        end;
      58:
        begin
          CastlingRemove('k');
          CastlingRemove('q');
        end;
    end;
  if activeColor * board[x1, y1] = cRook then
    case 10 * x1 + y1 of
      81: CastlingRemove('K');
      11: CastlingRemove('Q');
      88: CastlingRemove('k');
      18: CastlingRemove('q');
    end;
  if (activeColor * board[x1, y1] = cPawn) and ((y2 - y1) * activeColor = 2) then
    enPassant := SquareToStr(x1, y1 + activeColor)
  else
    enPassant := '-';
  if (activeColor * board[x1, y1] = cKing) and (Abs(x2 - x1) = 2) then
    case 10 * x2 + y2 of
      71:
        begin
          board[8, 1] := cNil;
          board[6, 1] := cWhite * cRook;
        end;
      31:
        begin
          board[1, 1] := cNil;
          board[4, 1] := cWhite * cRook;
        end;
      78:
        begin
          board[8, 8] := cNil;
          board[6, 8] := cBlack * cRook;
        end;
      38:
        begin
          board[1, 8] := cNil;
          board[4, 8] := cBlack * cRook;
        end;
    end;
  if (activeColor * board[x1, y1] = cPawn) and (x2 <> x1) and (board[x2, y2] = cNil) then
    board[x2, y1] := cNil;
  if (activeColor * board[x1, y1] = cPawn) or (board[x2, y2] <> cNil) then
    halfMoves := 0
  else
    Inc(halfMoves);
  if activeColor = cBlack then
    Inc(fullMove);
  if (activeColor * board[x1, y1] = cPawn) and ((y2 = 1) or (y2 = 8)) then
    board[x1, y1] := board[x1, y1] * aPromotion;
  board[x2, y2] := board[x1, y1];
  board[x1, y1] := cNil;
  activeColor := -1 * activeColor;
end;

function TPosition.ForsythEdwards: string;
const
  cSymbols: string = 'kqrbnp PNBRQK';
var
  x, y, z: integer;
  s: string;
begin
  result := '';
  x := 1;
  y := 8;
  z := 0;
  while y > 0 do
  begin
    //ToLog(Format('%d %d %d', [x, y, z]));
    if board[x, y] = cNil then
    begin
      while (x + z <= 8) and (board[x + z, y] = cNil) do
        Inc(z);
      result := Concat(result, Chr(Ord('0') + z));
      Inc(x, z);
      z := 0;
    end else
    begin
      result := Concat(result, cSymbols[board[x, y] + 7]);
      Inc(x);
    end;
    if x = 9 then
    begin
      if y > 1 then
        result := Concat(result, '/');
      x := 1;
      Dec(y);
    end;
  end;
  result := Concat(result, ' ');
  if activeColor = cWhite then
    s := 'w'
  else
    s := 'b';
  result := Concat(result, s);
  result := Concat(result, ' ');
  result := Concat(result, castling);
  result := Concat(result, ' ');
  result := Concat(result, enPassant);
  result := Concat(result, ' ');
  Str(halfMoves, s);
  result := Concat(result, s);
  result := Concat(result, ' ');
  Str(fullMove, s);
  result := Concat(result, s);
end;

procedure TPosition.UpdateVars;
begin
  GenerateMoves(-1 * activeColor, false);
  check1 := Check;
  castlingcheck1[e1g1] := CastlingCheck('K');
  castlingcheck1[e1c1] := CastlingCheck('Q');
  castlingcheck1[e8g8] := CastlingCheck('k');
  castlingcheck1[e8c8] := CastlingCheck('q');
end;

constructor TGame.Create;
begin
  inherited Create;
  current := TPosition.Create;
  next := TPosition.Create;
  list := TStringList.Create;
end;

constructor TGame.Create(const s: string);
begin
  inherited Create;
  current := TPosition.Create;
  next := TPosition.Create;
  list := TStringList.Create;
  current.Initialize(s);
end;

destructor TGame.Destroy;
begin
  current.Free;
  next.Free;
  list.Free;
  inherited Destroy;
end;

procedure TGame.GenerateLegalMoves;
var
  i: integer;
  s: string;
begin
  s := current.ForsythEdwards;
  current.UpdateVars;
  current.GenerateMoves(current.activeColor, true);
  list.Clear;
  for i := 0 to current.list.count - 1 do
  begin
    next.Initialize(s);
    next.PlayMove(current.list[i], cQueen);
    next.activeColor := -1 * next.activeColor;
    next.list.Clear;
    next.GenerateMoves(-1 * next.activeColor, false);
    next.UpdateVars;
    if not next.check1 then
      list.Add(current.list[i]);
  end;
  if list.count = 0 then
  begin
    if current.check1 then
      state := checkmate
    else
      state := stalemate;
  end else
    if current.halfMoves = 50 then
      state := draw
    else
      state := progress;
end;

function TGame.IsLegal(const s: string): boolean;
var
  i: integer;
begin
  list.Sort;
  result := list.Find(s, i);
end;

function TGame.ActiveColor: integer;
begin
  result := current.activeColor;
end;

function TGame.Check: boolean;
begin
  result := current.check1;
end;

function TGame.ForsythEdwards: string;
begin
  result := current.ForsythEdwards;
end;

procedure TGame.PlayMove(const aMove: string; const aPromotion: integer);
begin
  current.PlayMove(aMove, aPromotion);
end;

function TGame.CurrentBoardAsText: string;
begin
  result := current.BoardAsText;
end;

function TGame.IsPromotion(const aMove: string): boolean;
var
  x1, y1, {x2, }y2: integer;
begin
  x1 := Ord(aMove[1]) - Ord('a') + 1;
  y1 := Ord(aMove[2]) - Ord('1') + 1;
  //x2 := Ord(aMove[3]) - Ord('a') + 1;
  y2 := Ord(aMove[4]) - Ord('1') + 1;
  
  result := (Abs(current.board[x1, y1]) = cPawn) and (y2 in [1, 8]);
end;

function TGame.IsCastling(const aMove: string): string;
var
  x1, y1, x2, y2: integer;
begin
  x1 := Ord(aMove[1]) - Ord('a') + 1;
  y1 := Ord(aMove[2]) - Ord('1') + 1;
  x2 := Ord(aMove[3]) - Ord('a') + 1;
  y2 := Ord(aMove[4]) - Ord('1') + 1;
  
  result := '';
  if (Abs(current.board[x1, y1]) = cKing) and (Abs(x2 - x1) = 2) then
    case 10 * x2 + y2 of
      71: result := 'h1f1';
      31: result := 'a1d1';
      78: result := 'h8f8';
      38: result := 'a8d8';
    end;
end;

function TGame.IsEnPassant(const aMove: string): string;
var
  x1, y1, x2, y2: integer;
begin
  x1 := Ord(aMove[1]) - Ord('a') + 1;
  y1 := Ord(aMove[2]) - Ord('1') + 1;
  x2 := Ord(aMove[3]) - Ord('a') + 1;
  y2 := Ord(aMove[4]) - Ord('1') + 1;
  
  result := '';
  if (Abs(current.board[x1, y1]) = cPawn) and (Abs(current.board[x2, y2]) = cNil) and (x2 <> x1) then
    result := SquareToStr(x2, y1);
end;

var
  logdir: string;
  direxist: boolean;

initialization
  loglist := TStringList.Create;
  logfile := ChangeFileExt({$I %FILE%}, '.log');
  logdir := ExtractFilePath(ParamStr(0)) + 'log';
  direxist := DirectoryExists(logdir) or CreateDir(logdir);
  
finalization
  with loglist do
  begin
    if direxist then SaveToFile(logdir + '\' + logfile);
    Free;
  end;

end.

